@@ -17,11 +17,12 @@
 #include <ext2_fs.h>
 #include <ext2fs.h>
 
-#include <asm/addrspace.h>
 #include "arcboot.h"
 
 #include <subarch.h>
 
+#define KSEG0ADDR(addr)	(((addr) & 0x1fffffff) | 0x80000000)
+
 #define ANSI_CLEAR	"\033[2J"
 #define CONF_FILE	"/etc/arcboot.conf"
 
@@ -125,7 +126,7 @@ void InitMalloc(void)
 
 int isEnvVar(const char* arg)
 {
-	int i;
+	unsigned int i;
 
 	for (i = 0; i < NENTS(env_vars); i++) {
 		if(strncmp( env_vars[i], arg, strlen(env_vars[i]) ) ==  0)
@@ -179,6 +180,126 @@ int ProcessArguments(LONG argc, CHAR * argv[])
 	return 0;
 }
 
+int LoadProgramSegments32(ext2_file_t file, Elf_Ehdr * header, void *segments)
+{
+	int idx;
+	int loaded = 0;
+	Elf32_Phdr* segment=(Elf32_Phdr*)segments;
+	errcode_t status;
+	size_t size;
+
+	printf("Loading 32-bit executable\n\r");
+
+	for (idx = 0; idx < header->header32.e_phnum; idx++) {
+		if (segment->p_type == PT_LOAD) {
+		    printf
+			("Loading program segment %u at 0x%x, offset=0x%x, size = 0x%x\n\r",
+		        idx + 1, KSEG0ADDR(segment->p_vaddr),
+		        segment->p_offset, segment->p_filesz);
+
+		    status =
+			ext2fs_file_lseek(file, segment->p_offset,
+				        EXT2_SEEK_SET, NULL);
+		    if (status != 0) {
+			    print_ext2fs_error(status);
+			    Fatal("Cannot seek to program segment\n\r");
+		    }
+
+		    arc_do_progress = 1;
+		    status = ext2fs_file_read(file,
+					  (void *) (KSEG0ADDR(
+					            segment->p_vaddr)),
+					  segment->p_filesz, NULL);
+		    printf("\n\n\r");	/* Clear progress */
+		    arc_do_progress = 0;
+		    if (status != 0) {
+			print_ext2fs_error(status);
+			Fatal("Cannot read program segment\n\r");
+		    }
+
+		    size = segment->p_memsz - segment->p_filesz;
+		    if (size > 0) {
+			    printf
+				("Zeroing memory at 0x%x, size = 0x%x\n\r",
+			        (KSEG0ADDR(segment->p_vaddr +
+			        segment->p_filesz)), size);
+			    memset((void *)
+			       (KSEG0ADDR(segment->
+				 p_vaddr + segment->p_filesz)), 0, size);
+		    }
+
+		    loaded = 1;
+	    }
+
+	    segment =
+		(Elf32_Phdr *) (((char *) segment) +
+			    header->header32.e_phentsize);
+    }
+
+	return loaded;
+}
+
+int LoadProgramSegments64(ext2_file_t file, Elf_Ehdr * header, void *segments)
+{
+	int idx;
+	int loaded = 0;
+ 	Elf64_Phdr* segment=(Elf64_Phdr*)segments;
+	errcode_t status;
+	size_t size;
+
+	is64=1;
+	printf("Loading 64-bit executable\n\r");
+
+	for (idx = 0; idx < header->header64.e_phnum; idx++) {
+		if (segment->p_type == PT_LOAD) {
+		    printf ("Loading program segment %u at 0x%x, "
+			    "offset=0x%lx %lx, size = 0x%lx %lx\n\r",
+			    idx + 1,
+			    (int)KSEG0ADDR(segment->p_vaddr),
+			    (long)(segment->p_offset>>32),
+			    (long)(segment->p_offset&0xffffffff),
+			    (long)(segment->p_filesz>>32),
+			    (long)(segment->p_filesz&0xffffffff));
+
+		    status =
+			ext2fs_file_lseek(file, segment->p_offset,
+				        EXT2_SEEK_SET, NULL);
+		    if (status != 0) {
+			    print_ext2fs_error(status);
+			    Fatal("Cannot seek to program segment\n\r");
+		    }
+
+		    arc_do_progress = 1;
+		    status = ext2fs_file_read(file,
+					  (void *) (KSEG0ADDR(
+					            segment->p_vaddr)),
+					  segment->p_filesz, NULL);
+		    arc_do_progress = 0;
+		    if (status != 0) {
+			print_ext2fs_error(status);
+			Fatal("Cannot read program segment\n\r");
+		    }
+
+		    size = segment->p_memsz - segment->p_filesz;
+		    if (size > 0) {
+			    printf
+				("Zeroing memory at 0x%lx, size = 0x%lx\n\r",
+			        (KSEG0ADDR(segment->p_vaddr +
+			        segment->p_filesz)), size);
+			    memset((void *)
+			       (KSEG0ADDR(segment->
+				 p_vaddr + segment->p_filesz)), 0, size);
+		    }
+
+		    loaded = 1;
+		}
+
+		segment = (Elf64_Phdr *) (((char *) segment)
+					  + header->header64.e_phentsize);
+	}
+
+	return loaded;
+}
 
 void LoadProgramSegments(ext2_file_t file, Elf_Ehdr * header)
 {
@@ -224,99 +345,10 @@ void LoadProgramSegments(ext2_file_t file, Elf_Ehdr * header)
 		Fatal("Cannot read program segment headers\n\r");
 	}
 
-	if(header->e_ident[EI_CLASS] == ELFCLASS32) {
-	    Elf32_Phdr* segment=(Elf32_Phdr*)segments;
-	    printf("Loading 32-bit executable\n\r");
-	    for (idx = 0; idx < header->header32.e_phnum; idx++) {
-		    if (segment->p_type == PT_LOAD) {
-			    printf
-				("Loading program segment %u at 0x%x, offset=0x%x, size = 0x%x\n\r",
-			        idx + 1, KSEG0ADDR(segment->p_vaddr),
-			        segment->p_offset, segment->p_filesz);
-
-			    status =
-				ext2fs_file_lseek(file, segment->p_offset,
-					        EXT2_SEEK_SET, NULL);
-			    if (status != 0) {
-				    print_ext2fs_error(status);
-				    Fatal("Cannot seek to program segment\n\r");
-			    }
-
-			    status = ext2fs_file_read(file,
-						  (void *) (KSEG0ADDR(
-						            segment->p_vaddr)),
-						  segment->p_filesz, NULL);
-			    if (status != 0) {
-				print_ext2fs_error(status);
-				Fatal("Cannot read program segment\n\r");
-			    }
-
-			    size = segment->p_memsz - segment->p_filesz;
-			    if (size > 0) {
-				    printf
-					("Zeroing memory at 0x%x, size = 0x%x\n\r",
-				        (KSEG0ADDR(segment->p_vaddr +
-				        segment->p_filesz)), size);
-				    memset((void *)
-				       (KSEG0ADDR(segment->
-					 p_vaddr + segment->p_filesz)), 0, size);
-			    }
-
-			    loaded = True;
-		    }
-
-		    segment =
-			(Elf32_Phdr *) (((char *) segment) +
-				    header->header32.e_phentsize);
-	    }
-	} else {
-	    Elf64_Phdr* segment=(Elf64_Phdr*)segments;
-	    is64=1;
-	    printf("Loading 64-bit executable\n\r");
-	    for (idx = 0; idx < header->header64.e_phnum; idx++) {
-		    if (segment->p_type == PT_LOAD) {
-			    printf
-				("Loading program segment %u at 0x%x, offset=0x%lx %lx, size = 0x%lx %lx\n\r",
-			        idx + 1, KSEG0ADDR(segment->p_vaddr),
-			        (long)(segment->p_offset>>32),(long)(segment->p_offset&0xffffffff),
-				 (long)(segment->p_filesz>>32),(long)(segment->p_filesz&0xffffffff));
-
-			    status =
-				ext2fs_file_lseek(file, segment->p_offset,
-					        EXT2_SEEK_SET, NULL);
-			    if (status != 0) {
-				    print_ext2fs_error(status);
-				    Fatal("Cannot seek to program segment\n\r");
-			    }
-
-			    status = ext2fs_file_read(file,
-						  (void *) (KSEG0ADDR(
-						            segment->p_vaddr)),
-						  segment->p_filesz, NULL);
-			    if (status != 0) {
-				print_ext2fs_error(status);
-				Fatal("Cannot read program segment\n\r");
-			    }
-
-			    size = segment->p_memsz - segment->p_filesz;
-			    if (size > 0) {
-				    printf
-					("Zeroing memory at 0x%x, size = 0x%x\n\r",
-				        (KSEG0ADDR(segment->p_vaddr +
-				        segment->p_filesz)), size);
-				    memset((void *)
-				       (KSEG0ADDR(segment->
-					 p_vaddr + segment->p_filesz)), 0, size);
-			    }
-
-			    loaded = True;
-		    }
-
-		    segment =
-			(Elf64_Phdr *) (((char *) segment) +
-				    header->header64.e_phentsize);
-	    }
-	}
+	if(header->e_ident[EI_CLASS] == ELFCLASS32)
+		loaded = LoadProgramSegments32(file, header, segments);
+	else
+		loaded = LoadProgramSegments64(file, header, segments);
 
 	if (!loaded)
 		Fatal("No loadable program segments found\n\r");
@@ -431,13 +463,15 @@ void printCmdLine(int argc, CHAR *argv[])
 void _start64(LONG argc, CHAR * argv[], CHAR * envp[],
 	      unsigned long long *addr)
 {
-	asm volatile (".set mips4\n"
-		      "\t.set noreorder\n"
-		      "\tld $2, 0($7)\n"
-		      "\tj $2\n"
-		      "\tnop\n"
-		      "\t.set noreorder\n"
-		      "\t.set mips0");
+	__asm__ __volatile__(
+		".set push\n"
+		"\t.set mips3\n"
+		"\t.set noreorder\n"
+		"\t.set noat\n"
+		"\tld $1, 0($7)\n"
+		"\tjr $1\n"
+		"\t nop\n"
+		"\t.set pop");
 }
 
 void _start(LONG argc, CHAR *argv[], CHAR *envp[])
@@ -465,10 +499,12 @@ void _start(LONG argc, CHAR *argv[], CHAR *envp[])
 	    OSLoadFilename = argv[1];
 	}
 
-	if (OSLoadPartition == NULL)
-		Fatal("Invalid load parition\n\r");
+	/* Fall back to "Linux" as default name. */
 	if (OSLoadFilename == NULL)
-		Fatal("Invalid load filename\n\r");
+		OSLoadFilename = "Linux";
+
+	if (OSLoadPartition == NULL)
+		Fatal("Invalid load partition\n\r");
 #if DEBUG
 	printf("OSLoadPartition: %s\n\r", OSLoadPartition);
 	printf("OSLoadFilename: %s\n\r", OSLoadFilename);
@@ -509,14 +545,14 @@ void _start(LONG argc, CHAR *argv[], CHAR *envp[])
 #endif
 	if( kernel_entry64 ) {
 	    if(is64==0){
-		printf("Starting 32-bit kernel\n\r");
+		printf("Starting ELF32 kernel\n\r");
 		ArcFlushAllCaches();
 		((void (*)(int argc, CHAR * argv[], CHAR * envp[]))
 		    kernel_entry32)(nargc ,nargv, envp);
 	    } else {
-		printf("Starting 64-bit kernel\n\r");
+		printf("Starting ELF64 kernel\n\r");
 		ArcFlushAllCaches();
-		_start64(nargc ,nargv, envp, &kernel_entry64);
+		_start64(nargc, nargv, envp, &kernel_entry64);
 	    }
 	} else {
 		printf("Invalid kernel entry NULL\n\r");
